@@ -219,6 +219,8 @@ fwupd_device_flag_to_string(FwupdDeviceFlags device_flag)
 		return "emulation-tag";
 	if (device_flag == FWUPD_DEVICE_FLAG_ONLY_EXPLICIT_UPDATES)
 		return "only-explicit-updates";
+	if (device_flag == FWUPD_DEVICE_FLAG_INSTALL_SKIP_VERSION_CHECK)
+		return "install-skip-version-check";
 	if (device_flag == FWUPD_DEVICE_FLAG_UNKNOWN)
 		return "unknown";
 	return NULL;
@@ -344,6 +346,8 @@ fwupd_device_flag_from_string(const gchar *device_flag)
 		return FWUPD_DEVICE_FLAG_EMULATION_TAG;
 	if (g_strcmp0(device_flag, "only-explicit-updates") == 0)
 		return FWUPD_DEVICE_FLAG_ONLY_EXPLICIT_UPDATES;
+	if (g_strcmp0(device_flag, "install-skip-version-check") == 0)
+		return FWUPD_DEVICE_FLAG_INSTALL_SKIP_VERSION_CHECK;
 	return FWUPD_DEVICE_FLAG_UNKNOWN;
 }
 
@@ -926,6 +930,8 @@ fwupd_version_format_from_string(const gchar *str)
 		return FWUPD_VERSION_FORMAT_DELL_BIOS;
 	if (g_strcmp0(str, "hex") == 0)
 		return FWUPD_VERSION_FORMAT_HEX;
+	if (g_strcmp0(str, "dell-bios-msb") == 0)
+		return FWUPD_VERSION_FORMAT_DELL_BIOS_MSB;
 	return FWUPD_VERSION_FORMAT_UNKNOWN;
 }
 
@@ -966,5 +972,7 @@ fwupd_version_format_to_string(FwupdVersionFormat kind)
 		return "dell-bios";
 	if (kind == FWUPD_VERSION_FORMAT_HEX)
 		return "hex";
+	if (kind == FWUPD_VERSION_FORMAT_DELL_BIOS_MSB)
+		return "dell-bios-msb";
 	return NULL;
 }
