@@ -14,6 +14,7 @@
 
 #include "fu-byte-array.h"
 #include "fu-common.h"
+#include "fu-coswid-common.h"
 #include "fu-coswid-firmware.h"
 #include "fu-coswid-struct.h"
 
@@ -38,10 +39,12 @@ typedef struct {
 G_DEFINE_TYPE_WITH_PRIVATE(FuCoswidFirmware, fu_coswid_firmware, FU_TYPE_FIRMWARE)
 #define GET_PRIVATE(o) (fu_coswid_firmware_get_instance_private(o))
 
+#define FU_COSWID_FIRMWARE_MAX_ALLOCATION 0x32000
+
 typedef struct {
 	gchar *name;
 	gchar *regid;
-	FuCoswidEntityRole roles[6];
+	guint8 roles; /* bitfield of FuCoswidEntityRole */
 } FuCoswidFirmwareEntity;
 
 typedef struct {
@@ -109,62 +112,33 @@ fu_coswid_firmware_payload_free(FuCoswidFirmwarePayload *payload)
 G_DEFINE_AUTOPTR_CLEANUP_FUNC(FuCoswidFirmwarePayload, fu_coswid_firmware_payload_free)
 
 #ifdef HAVE_CBOR
-G_DEFINE_AUTOPTR_CLEANUP_FUNC(cbor_item_t, cbor_intermediate_decref)
-
-static gchar *
-fu_coswid_firmware_strndup(cbor_item_t *item)
-{
-	if (!cbor_string_is_definite(item))
-		return NULL;
-	return g_strndup((const gchar *)cbor_string_handle(item), cbor_string_length(item));
-}
-
-typedef gboolean (*FuCoswidFirmwareItemFunc)(FuCoswidFirmware *self,
-					     cbor_item_t *item,
-					     gpointer user_data,
-					     GError **error) G_GNUC_WARN_UNUSED_RESULT;
 
+/* @userdata: a #FuCoswidFirmware */
 static gboolean
-fu_coswid_firmware_parse_one_or_many(FuCoswidFirmware *self,
-				     cbor_item_t *item,
-				     FuCoswidFirmwareItemFunc func,
-				     gpointer user_data,
-				     GError **error)
-{
-	/* one */
-	if (cbor_isa_map(item))
-		return func(self, item, user_data, error);
-
-	/* many */
-	if (cbor_isa_array(item)) {
-		for (guint j = 0; j < cbor_array_size(item); j++) {
-			g_autoptr(cbor_item_t) value = cbor_array_get(item, j);
-			if (!func(self, value, user_data, error))
-				return FALSE;
-		}
-		return TRUE;
-	}
-
-	/* not sure what to do */
-	g_set_error_literal(error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA, "neither an array or map");
-	return FALSE;
-}
-
-static gboolean
-fu_coswid_firmware_parse_meta(FuCoswidFirmware *self,
-			      cbor_item_t *item,
-			      gpointer user_data,
-			      GError **error)
+fu_coswid_firmware_parse_meta(cbor_item_t *item, gpointer user_data, GError **error)
 {
+	FuCoswidFirmware *self = FU_COSWID_FIRMWARE(user_data);
 	FuCoswidFirmwarePrivate *priv = GET_PRIVATE(self);
 	struct cbor_pair *pairs = cbor_map_handle(item);
 
 	for (gsize i = 0; i < cbor_map_size(item); i++) {
-		FuCoswidTag tag_id = cbor_get_uint8(pairs[i].key);
+		FuCoswidTag tag_id = 0;
+		if (!fu_coswid_read_tag(pairs[i].key, &tag_id, error))
+			return FALSE;
 		if (tag_id == FU_COSWID_TAG_SUMMARY) {
-			priv->summary = fu_coswid_firmware_strndup(pairs[i].value);
+			g_free(priv->summary);
+			priv->summary = fu_coswid_read_string(pairs[i].value, error);
+			if (priv->summary == NULL) {
+				g_prefix_error(error, "failed to parse summary: ");
+				return FALSE;
+			}
 		} else if (tag_id == FU_COSWID_TAG_COLLOQUIAL_VERSION) {
-			priv->colloquial_version = fu_coswid_firmware_strndup(pairs[i].value);
+			g_free(priv->colloquial_version);
+			priv->colloquial_version = fu_coswid_read_string(pairs[i].value, error);
+			if (priv->colloquial_version == NULL) {
+				g_prefix_error(error, "failed to parse colloquial-version: ");
+				return FALSE;
+			}
 		} else {
 			g_debug("unhandled tag %s from %s",
 				fu_coswid_tag_to_string(tag_id),
@@ -176,25 +150,33 @@ fu_coswid_firmware_parse_meta(FuCoswidFirmware *self,
 	return TRUE;
 }
 
+/* @userdata: a #FuCoswidFirmware */
 static gboolean
-fu_coswid_firmware_parse_link(FuCoswidFirmware *self,
-			      cbor_item_t *item,
-			      gpointer user_data,
-			      GError **error)
+fu_coswid_firmware_parse_link(cbor_item_t *item, gpointer user_data, GError **error)
 {
+	FuCoswidFirmware *self = FU_COSWID_FIRMWARE(user_data);
 	FuCoswidFirmwarePrivate *priv = GET_PRIVATE(self);
 	struct cbor_pair *pairs = cbor_map_handle(item);
 	g_autoptr(FuCoswidFirmwareLink) link = g_new0(FuCoswidFirmwareLink, 1);
 
 	for (gsize i = 0; i < cbor_map_size(item); i++) {
-		FuCoswidTag tag_id = cbor_get_uint8(pairs[i].key);
+		FuCoswidTag tag_id = 0;
+		if (!fu_coswid_read_tag(pairs[i].key, &tag_id, error))
+			return FALSE;
 		if (tag_id == FU_COSWID_TAG_HREF) {
-			link->href = fu_coswid_firmware_strndup(pairs[i].value);
+			g_free(link->href);
+			link->href = fu_coswid_read_string(pairs[i].value, error);
+			if (link->href == NULL) {
+				g_prefix_error(error, "failed to parse link href: ");
+				return FALSE;
+			}
 		} else if (tag_id == FU_COSWID_TAG_REL) {
-			if (cbor_isa_negint(pairs[i].value))
-				link->rel = (-1) - cbor_get_uint8(pairs[i].value);
-			else
-				link->rel = cbor_get_uint8(pairs[i].value);
+			gint8 tmp = 0;
+			if (!fu_coswid_read_s8(pairs[i].value, &tmp, error)) {
+				g_prefix_error(error, "failed to parse link rel: ");
+				return FALSE;
+			}
+			link->rel = tmp;
 		} else {
 			g_debug("unhandled tag %s from %s",
 				fu_coswid_tag_to_string(tag_id),
@@ -207,18 +189,33 @@ fu_coswid_firmware_parse_link(FuCoswidFirmware *self,
 	return TRUE;
 }
 
+/* @userdata: a #FuCoswidFirmwarePayload */
 static gboolean
-fu_coswid_firmware_parse_hash(FuCoswidFirmware *self,
-			      cbor_item_t *item,
-			      gpointer user_data,
-			      GError **error)
+fu_coswid_firmware_parse_hash(cbor_item_t *item, gpointer user_data, GError **error)
 {
 	FuCoswidFirmwarePayload *payload = (FuCoswidFirmwarePayload *)user_data;
+	guint8 alg_id8 = 0;
 	g_autoptr(FuCoswidFirmwareHash) hash = g_new0(FuCoswidFirmwareHash, 1);
-	g_autoptr(cbor_item_t) hash_item_alg_id = cbor_array_get(item, 0);
-	g_autoptr(cbor_item_t) hash_item_value = cbor_array_get(item, 1);
+	g_autoptr(cbor_item_t) hash_item_alg_id = NULL;
+	g_autoptr(cbor_item_t) hash_item_value = NULL;
 
 	/* sanity check */
+	if (!cbor_isa_array(item)) {
+		g_set_error_literal(error,
+				    G_IO_ERROR,
+				    G_IO_ERROR_INVALID_DATA,
+				    "hash item is not an array");
+		return FALSE;
+	}
+	if (cbor_array_size(item) != 2) {
+		g_set_error_literal(error,
+				    G_IO_ERROR,
+				    G_IO_ERROR_INVALID_DATA,
+				    "hash array has invalid size");
+		return FALSE;
+	}
+	hash_item_alg_id = cbor_array_get(item, 0);
+	hash_item_value = cbor_array_get(item, 1);
 	if (hash_item_alg_id == NULL || hash_item_value == NULL) {
 		g_set_error_literal(error,
 				    G_IO_ERROR,
@@ -226,62 +223,70 @@ fu_coswid_firmware_parse_hash(FuCoswidFirmware *self,
 				    "invalid hash item");
 		return FALSE;
 	}
+	if (!fu_coswid_read_u8(hash_item_alg_id, &alg_id8, error)) {
+		g_prefix_error(error, "failed to parse hash alg-id: ");
+		return FALSE;
+	}
 
 	/* success */
-	hash->alg_id = cbor_get_uint8(hash_item_alg_id);
-	hash->value = g_byte_array_new();
-	g_byte_array_append(hash->value,
-			    cbor_bytestring_handle(hash_item_value),
-			    cbor_bytestring_length(hash_item_value));
+	hash->alg_id = alg_id8;
+	hash->value = fu_coswid_read_byte_array(hash_item_value, error);
+	if (hash->value == NULL) {
+		g_prefix_error(error, "failed to parse hash value: ");
+		return FALSE;
+	}
 	g_ptr_array_add(payload->hashes, g_steal_pointer(&hash));
 	return TRUE;
 }
 
+/* @userdata: a #FuCoswidFirmwarePayload */
 static gboolean
-fu_coswid_firmware_parse_hash_array(FuCoswidFirmware *self,
-				    cbor_item_t *item,
-				    gpointer user_data,
-				    GError **error)
+fu_coswid_firmware_parse_hash_array(cbor_item_t *item, gpointer user_data, GError **error)
 {
 	for (guint j = 0; j < cbor_array_size(item); j++) {
 		g_autoptr(cbor_item_t) value = cbor_array_get(item, j);
-		if (!fu_coswid_firmware_parse_hash(self, value, user_data, error))
+		if (!fu_coswid_firmware_parse_hash(value, user_data, error))
 			return FALSE;
 	}
 	return TRUE;
 }
 
+/* @userdata: a #FuCoswidFirmware */
 static gboolean
-fu_coswid_firmware_parse_file(FuCoswidFirmware *self,
-			      cbor_item_t *item,
-			      gpointer user_data,
-			      GError **error)
+fu_coswid_firmware_parse_file(cbor_item_t *item, gpointer user_data, GError **error)
 {
+	FuCoswidFirmware *self = FU_COSWID_FIRMWARE(user_data);
 	FuCoswidFirmwarePrivate *priv = GET_PRIVATE(self);
 	struct cbor_pair *pairs = cbor_map_handle(item);
 	g_autoptr(FuCoswidFirmwarePayload) payload = fu_coswid_firmware_payload_new();
 
 	for (gsize i = 0; i < cbor_map_size(item); i++) {
-		FuCoswidTag tag_id = cbor_get_uint8(pairs[i].key);
+		FuCoswidTag tag_id = 0;
+		if (!fu_coswid_read_tag(pairs[i].key, &tag_id, error))
+			return FALSE;
 		if (tag_id == FU_COSWID_TAG_FS_NAME) {
-			payload->name = fu_coswid_firmware_strndup(pairs[i].value);
+			g_free(payload->name);
+			payload->name = fu_coswid_read_string(pairs[i].value, error);
+			if (payload->name == NULL) {
+				g_prefix_error(error, "failed to parse payload name: ");
+				return FALSE;
+			}
 		} else if (tag_id == FU_COSWID_TAG_SIZE) {
-			payload->size = cbor_get_int(pairs[i].value);
+			if (!fu_coswid_read_u64(pairs[i].value, &payload->size, error))
+				return FALSE;
 		} else if (tag_id == FU_COSWID_TAG_HASH) {
 			if (cbor_isa_array(pairs[i].value) &&
 			    cbor_array_size(pairs[i].value) >= 1) {
 				g_autoptr(cbor_item_t) value = cbor_array_get(pairs[i].value, 0);
-				/* we can't use fu_coswid_firmware_parse_one_or_many() here as
+				/* we can't use fu_coswid_parse_one_or_many() here as
 				 * the hash is an array, not a map -- for some reason */
 				if (cbor_isa_array(value)) {
-					if (!fu_coswid_firmware_parse_hash_array(self,
-										 pairs[i].value,
+					if (!fu_coswid_firmware_parse_hash_array(pairs[i].value,
 										 payload,
 										 error))
 						return FALSE;
 				} else {
-					if (!fu_coswid_firmware_parse_hash(self,
-									   pairs[i].value,
+					if (!fu_coswid_firmware_parse_hash(pairs[i].value,
 									   payload,
 									   error))
 						return FALSE;
@@ -305,21 +310,21 @@ fu_coswid_firmware_parse_file(FuCoswidFirmware *self,
 	return TRUE;
 }
 
+/* @userdata: a #FuCoswidFirmware */
 static gboolean
-fu_coswid_firmware_parse_path_elements(FuCoswidFirmware *self,
-				       cbor_item_t *item,
-				       gpointer user_data,
-				       GError **error)
+fu_coswid_firmware_parse_path_elements(cbor_item_t *item, gpointer user_data, GError **error)
 {
+	FuCoswidFirmware *self = FU_COSWID_FIRMWARE(user_data);
 	struct cbor_pair *pairs = cbor_map_handle(item);
 	for (gsize i = 0; i < cbor_map_size(item); i++) {
-		FuCoswidTag tag_id = cbor_get_uint8(pairs[i].key);
+		FuCoswidTag tag_id = 0;
+		if (!fu_coswid_read_tag(pairs[i].key, &tag_id, error))
+			return FALSE;
 		if (tag_id == FU_COSWID_TAG_FILE) {
-			if (!fu_coswid_firmware_parse_one_or_many(self,
-								  pairs[i].value,
-								  fu_coswid_firmware_parse_file,
-								  NULL, /* user_data */
-								  error))
+			if (!fu_coswid_parse_one_or_many(pairs[i].value,
+							 fu_coswid_firmware_parse_file,
+							 self, /* user_data */
+							 error))
 				return FALSE;
 		} else {
 			g_debug("unhandled tag %s from %s",
@@ -332,22 +337,21 @@ fu_coswid_firmware_parse_path_elements(FuCoswidFirmware *self,
 	return TRUE;
 }
 
+/* @userdata: a #FuCoswidFirmware */
 static gboolean
-fu_coswid_firmware_parse_directory(FuCoswidFirmware *self,
-				   cbor_item_t *item,
-				   gpointer user_data,
-				   GError **error)
+fu_coswid_firmware_parse_directory(cbor_item_t *item, gpointer user_data, GError **error)
 {
+	FuCoswidFirmware *self = FU_COSWID_FIRMWARE(user_data);
 	struct cbor_pair *pairs = cbor_map_handle(item);
 	for (gsize i = 0; i < cbor_map_size(item); i++) {
-		FuCoswidTag tag_id = cbor_get_uint8(pairs[i].key);
+		FuCoswidTag tag_id = 0;
+		if (!fu_coswid_read_tag(pairs[i].key, &tag_id, error))
+			return FALSE;
 		if (tag_id == FU_COSWID_TAG_PATH_ELEMENTS) {
-			if (!fu_coswid_firmware_parse_one_or_many(
-				self,
-				pairs[i].value,
-				fu_coswid_firmware_parse_path_elements,
-				NULL, /* user_data */
-				error))
+			if (!fu_coswid_parse_one_or_many(pairs[i].value,
+							 fu_coswid_firmware_parse_path_elements,
+							 self, /* user_data */
+							 error))
 				return FALSE;
 		} else {
 			g_debug("unhandled tag %s from %s",
@@ -360,29 +364,27 @@ fu_coswid_firmware_parse_directory(FuCoswidFirmware *self,
 	return TRUE;
 }
 
+/* @userdata: a #FuCoswidFirmware */
 static gboolean
-fu_coswid_firmware_parse_payload(FuCoswidFirmware *self,
-				 cbor_item_t *item,
-				 gpointer user_data,
-				 GError **error)
+fu_coswid_firmware_parse_payload(cbor_item_t *item, gpointer user_data, GError **error)
 {
+	FuCoswidFirmware *self = FU_COSWID_FIRMWARE(user_data);
 	struct cbor_pair *pairs = cbor_map_handle(item);
 	for (gsize i = 0; i < cbor_map_size(item); i++) {
-		FuCoswidTag tag_id = cbor_get_uint8(pairs[i].key);
+		FuCoswidTag tag_id = 0;
+		if (!fu_coswid_read_tag(pairs[i].key, &tag_id, error))
+			return FALSE;
 		if (tag_id == FU_COSWID_TAG_FILE) {
-			if (!fu_coswid_firmware_parse_one_or_many(self,
-								  pairs[i].value,
-								  fu_coswid_firmware_parse_file,
-								  NULL, /* user_data */
-								  error))
+			if (!fu_coswid_parse_one_or_many(pairs[i].value,
+							 fu_coswid_firmware_parse_file,
+							 self, /* user_data */
+							 error))
 				return FALSE;
 		} else if (tag_id == FU_COSWID_TAG_DIRECTORY) {
-			if (!fu_coswid_firmware_parse_one_or_many(
-				self,
-				pairs[i].value,
-				fu_coswid_firmware_parse_directory,
-				NULL, /* user_data */
-				error))
+			if (!fu_coswid_parse_one_or_many(pairs[i].value,
+							 fu_coswid_firmware_parse_directory,
+							 self, /* user_data */
+							 error))
 				return FALSE;
 		} else {
 			g_debug("unhandled tag %s from %s",
@@ -395,39 +397,75 @@ fu_coswid_firmware_parse_payload(FuCoswidFirmware *self,
 	return TRUE;
 }
 
+/* @userdata: a #FuCoswidFirmware */
 static gboolean
-fu_coswid_firmware_parse_entity(FuCoswidFirmware *self,
-				cbor_item_t *item,
-				gpointer user_data,
-				GError **error)
+fu_coswid_firmware_parse_entity(cbor_item_t *item, gpointer user_data, GError **error)
 {
+	FuCoswidFirmware *self = FU_COSWID_FIRMWARE(user_data);
 	FuCoswidFirmwarePrivate *priv = GET_PRIVATE(self);
 	struct cbor_pair *pairs = cbor_map_handle(item);
-	guint entity_role_cnt = 0;
 	g_autoptr(FuCoswidFirmwareEntity) entity = g_new0(FuCoswidFirmwareEntity, 1);
 
 	for (gsize i = 0; i < cbor_map_size(item); i++) {
-		FuCoswidTag tag_id = cbor_get_uint8(pairs[i].key);
+		FuCoswidTag tag_id = 0;
+		if (!fu_coswid_read_tag(pairs[i].key, &tag_id, error))
+			return FALSE;
 		if (tag_id == FU_COSWID_TAG_ENTITY_NAME) {
-			entity->name = fu_coswid_firmware_strndup(pairs[i].value);
+			g_free(entity->name);
+			entity->name = fu_coswid_read_string(pairs[i].value, error);
+			if (entity->name == NULL) {
+				g_prefix_error(error, "failed to parse entity name: ");
+				return FALSE;
+			}
 		} else if (tag_id == FU_COSWID_TAG_REG_ID) {
-			entity->regid = fu_coswid_firmware_strndup(pairs[i].value);
+			g_free(entity->regid);
+			entity->regid = fu_coswid_read_string(pairs[i].value, error);
+			if (entity->regid == NULL) {
+				g_prefix_error(error, "failed to parse entity regid: ");
+				return FALSE;
+			}
 		} else if (tag_id == FU_COSWID_TAG_ROLE) {
 			if (cbor_isa_uint(pairs[i].value)) {
-				FuCoswidEntityRole role = cbor_get_uint8(pairs[i].value);
-				entity->roles[entity_role_cnt++] = role;
-			}
-			for (guint j = 0; j < cbor_array_size(pairs[i].value); j++) {
-				g_autoptr(cbor_item_t) value = cbor_array_get(pairs[i].value, j);
-				FuCoswidEntityRole role = cbor_get_uint8(value);
-				if (entity_role_cnt >= G_N_ELEMENTS(entity->roles)) {
-					g_set_error_literal(error,
-							    G_IO_ERROR,
-							    G_IO_ERROR_INVALID_DATA,
-							    "too many roles");
+				guint8 role8 = 0;
+				if (!fu_coswid_read_u8(pairs[i].value, &role8, error)) {
+					g_prefix_error(error, "failed to parse entity role: ");
 					return FALSE;
 				}
-				entity->roles[entity_role_cnt++] = role;
+				if (role8 >= FU_COSWID_ENTITY_ROLE_LAST) {
+					g_set_error(error,
+						    G_IO_ERROR,
+						    G_IO_ERROR_INVALID_DATA,
+						    "invalid entity role 0x%x",
+						    role8);
+					return FALSE;
+				}
+				entity->roles |= 1u << role8;
+			} else if (cbor_isa_array(pairs[i].value)) {
+				for (guint j = 0; j < cbor_array_size(pairs[i].value); j++) {
+					guint8 role8 = 0;
+					g_autoptr(cbor_item_t) value =
+					    cbor_array_get(pairs[i].value, j);
+					if (!fu_coswid_read_u8(value, &role8, error)) {
+						g_prefix_error(error,
+							       "failed to parse entity role: ");
+						return FALSE;
+					}
+					if (role8 >= FU_COSWID_ENTITY_ROLE_LAST) {
+						g_set_error(error,
+							    G_IO_ERROR,
+							    G_IO_ERROR_INVALID_DATA,
+							    "invalid entity role 0x%x",
+							    role8);
+						return FALSE;
+					}
+					entity->roles |= 1u << role8;
+				}
+			} else {
+				g_set_error_literal(error,
+						    G_IO_ERROR,
+						    G_IO_ERROR_INVALID_DATA,
+						    "entity role item is not an uint or array");
+				return FALSE;
 			}
 		} else {
 			g_debug("unhandled tag %s from %s",
@@ -442,6 +480,36 @@ fu_coswid_firmware_parse_entity(FuCoswidFirmware *self,
 }
 #endif
 
+#ifdef HAVE_CBOR_SET_ALLOCS
+static void *
+fu_coswid_firmware_malloc(size_t size)
+{
+	if (size > FU_COSWID_FIRMWARE_MAX_ALLOCATION) {
+		g_debug("failing CBOR allocation of 0x%x bytes", (guint)size);
+		return NULL;
+	}
+	/* libcbor expects a valid pointer for a zero sized allocation */
+	return g_malloc0(MAX(size, 1));
+}
+
+static void *
+fu_coswid_firmware_realloc(void *ptr, size_t size)
+{
+	if (size > FU_COSWID_FIRMWARE_MAX_ALLOCATION) {
+		g_debug("failing CBOR reallocation of 0x%x bytes", (guint)size);
+		return NULL;
+	}
+	/* libcbor expects a valid pointer for a zero sized allocation */
+	return g_realloc(ptr, MAX(size, 1));
+}
+
+static void
+fu_coswid_firmware_free(void *ptr)
+{
+	g_free(ptr);
+}
+#endif
+
 static gboolean
 fu_coswid_firmware_parse(FuFirmware *firmware,
 			 GBytes *fw,
@@ -486,55 +554,60 @@ fu_coswid_firmware_parse(FuFirmware *firmware,
 	/* parse out anything interesting */
 	pairs = cbor_map_handle(item);
 	for (gsize i = 0; i < cbor_map_size(item); i++) {
-		FuCoswidTag tag_id = cbor_get_uint8(pairs[i].key);
+		FuCoswidTag tag_id = 0;
+		if (!fu_coswid_read_tag(pairs[i].key, &tag_id, error))
+			return FALSE;
 
 		/* identity can be specified as a string or in binary */
 		if (tag_id == FU_COSWID_TAG_TAG_ID) {
-			g_autofree gchar *str = NULL;
-			if (cbor_isa_string(pairs[i].value)) {
-				str = fu_coswid_firmware_strndup(pairs[i].value);
-			} else if (cbor_isa_bytestring(pairs[i].value) &&
-				   cbor_bytestring_length(pairs[i].value) == 16) {
-				str = fwupd_guid_to_string(
-				    (const fwupd_guid_t *)cbor_bytestring_handle(pairs[i].value),
-				    FWUPD_GUID_FLAG_NONE);
+			g_autofree gchar *str = fu_coswid_read_string(pairs[i].value, error);
+			if (str == NULL) {
+				g_prefix_error(error, "failed to parse tag-id: ");
+				return FALSE;
 			}
-			if (str != NULL)
-				fu_firmware_set_id(firmware, str);
+			fu_firmware_set_id(firmware, str);
 		} else if (tag_id == FU_COSWID_TAG_SOFTWARE_NAME) {
-			priv->product = fu_coswid_firmware_strndup(pairs[i].value);
+			g_free(priv->product);
+			priv->product = fu_coswid_read_string(pairs[i].value, error);
+			if (priv->product == NULL) {
+				g_prefix_error(error, "failed to parse product: ");
+				return FALSE;
+			}
 		} else if (tag_id == FU_COSWID_TAG_SOFTWARE_VERSION) {
-			g_autofree gchar *str = fu_coswid_firmware_strndup(pairs[i].value);
+			g_autofree gchar *str = fu_coswid_read_string(pairs[i].value, error);
+			if (str == NULL) {
+				g_prefix_error(error, "failed to parse software-version: ");
+				return FALSE;
+			}
 			fu_firmware_set_version(firmware, str);
 		} else if (tag_id == FU_COSWID_TAG_VERSION_SCHEME) {
-			priv->version_scheme = cbor_get_uint16(pairs[i].value);
+			if (!fu_coswid_read_version_scheme(pairs[i].value,
+							   &priv->version_scheme,
+							   error))
+				return FALSE;
 		} else if (tag_id == FU_COSWID_TAG_SOFTWARE_META) {
-			if (!fu_coswid_firmware_parse_one_or_many(self,
-								  pairs[i].value,
-								  fu_coswid_firmware_parse_meta,
-								  NULL, /* user_data */
-								  error))
+			if (!fu_coswid_parse_one_or_many(pairs[i].value,
+							 fu_coswid_firmware_parse_meta,
+							 self, /* user_data */
+							 error))
 				return FALSE;
 		} else if (tag_id == FU_COSWID_TAG_LINK) {
-			if (!fu_coswid_firmware_parse_one_or_many(self,
-								  pairs[i].value,
-								  fu_coswid_firmware_parse_link,
-								  NULL, /* user_data */
-								  error))
+			if (!fu_coswid_parse_one_or_many(pairs[i].value,
+							 fu_coswid_firmware_parse_link,
+							 self, /* user_data */
+							 error))
 				return FALSE;
 		} else if (tag_id == FU_COSWID_TAG_PAYLOAD) {
-			if (!fu_coswid_firmware_parse_one_or_many(self,
-								  pairs[i].value,
-								  fu_coswid_firmware_parse_payload,
-								  NULL, /* user_data */
-								  error))
+			if (!fu_coswid_parse_one_or_many(pairs[i].value,
+							 fu_coswid_firmware_parse_payload,
+							 self, /* user_data */
+							 error))
 				return FALSE;
 		} else if (tag_id == FU_COSWID_TAG_ENTITY) {
-			if (!fu_coswid_firmware_parse_one_or_many(self,
-								  pairs[i].value,
-								  fu_coswid_firmware_parse_entity,
-								  NULL, /* user_data */
-								  error))
+			if (!fu_coswid_parse_one_or_many(pairs[i].value,
+							 fu_coswid_firmware_parse_entity,
+							 self, /* user_data */
+							 error))
 				return FALSE;
 		} else {
 			g_debug("unhandled tag %s from root", fu_coswid_tag_to_string(tag_id));
@@ -547,7 +620,7 @@ fu_coswid_firmware_parse(FuFirmware *firmware,
 		g_set_error_literal(error,
 				    G_IO_ERROR,
 				    G_IO_ERROR_NOT_SUPPORTED,
-				    "not enough SBoM data");
+				    "not enough SBOM data");
 		return FALSE;
 	}
 
@@ -610,76 +683,6 @@ fu_coswid_firmware_get_checksum(FuFirmware *firmware, GChecksumType csum_kind, G
 }
 
 #ifdef HAVE_CBOR
-static void
-fu_coswid_firmware_write_tag_string(cbor_item_t *root, FuCoswidTag tag, const gchar *item)
-{
-	g_autoptr(cbor_item_t) key = cbor_build_uint8(tag);
-	g_autoptr(cbor_item_t) val = cbor_build_string(item);
-	if (!cbor_map_add(root, (struct cbor_pair){.key = key, .value = val}))
-		g_critical("failed to push to indefinite map");
-}
-
-static void
-fu_coswid_firmware_write_tag_bytestring(cbor_item_t *root,
-					FuCoswidTag tag,
-					const guint8 *buf,
-					gsize bufsz)
-{
-	g_autoptr(cbor_item_t) key = cbor_build_uint8(tag);
-	g_autoptr(cbor_item_t) val = cbor_build_bytestring((cbor_data)buf, bufsz);
-	if (!cbor_map_add(root, (struct cbor_pair){.key = key, .value = val}))
-		g_critical("failed to push to indefinite map");
-}
-
-static void
-fu_coswid_firmware_write_tag_bool(cbor_item_t *root, FuCoswidTag tag, gboolean item)
-{
-	g_autoptr(cbor_item_t) key = cbor_build_uint8(tag);
-	g_autoptr(cbor_item_t) val = cbor_build_bool(item);
-	if (!cbor_map_add(root, (struct cbor_pair){.key = key, .value = val}))
-		g_critical("failed to push to indefinite map");
-}
-
-static void
-fu_coswid_firmware_write_tag_uint16(cbor_item_t *root, FuCoswidTag tag, guint16 item)
-{
-	g_autoptr(cbor_item_t) key = cbor_build_uint8(tag);
-	g_autoptr(cbor_item_t) val = cbor_build_uint16(item);
-	if (!cbor_map_add(root, (struct cbor_pair){.key = key, .value = val}))
-		g_critical("failed to push to indefinite map");
-}
-
-static void
-fu_coswid_firmware_write_tag_uint64(cbor_item_t *root, FuCoswidTag tag, guint64 item)
-{
-	g_autoptr(cbor_item_t) key = cbor_build_uint8(tag);
-	g_autoptr(cbor_item_t) val = cbor_build_uint64(item);
-	if (!cbor_map_add(root, (struct cbor_pair){.key = key, .value = val}))
-		g_critical("failed to push to indefinite map");
-}
-
-static void
-fu_coswid_firmware_write_tag_int8(cbor_item_t *root, FuCoswidTag tag, gint8 item)
-{
-	g_autoptr(cbor_item_t) key = cbor_build_uint8(tag);
-	g_autoptr(cbor_item_t) val = cbor_new_int8();
-	if (item >= 0) {
-		cbor_set_uint8(val, item);
-	} else {
-		cbor_set_uint8(val, 0xFF - item);
-		cbor_mark_negint(val);
-	}
-	if (!cbor_map_add(root, (struct cbor_pair){.key = key, .value = val}))
-		g_critical("failed to push to indefinite map");
-}
-
-static void
-fu_coswid_firmware_write_tag_item(cbor_item_t *root, FuCoswidTag tag, cbor_item_t *item)
-{
-	g_autoptr(cbor_item_t) key = cbor_build_uint8(tag);
-	if (!cbor_map_add(root, (struct cbor_pair){.key = key, .value = item}))
-		g_critical("failed to push to indefinite map");
-}
 
 static void
 fu_coswid_firmware_write_hash(cbor_item_t *root, FuCoswidFirmwareHash *hash)
@@ -702,12 +705,10 @@ fu_coswid_firmware_write_payload(cbor_item_t *root, FuCoswidFirmwarePayload *pay
 	g_autoptr(cbor_item_t) item_payload = cbor_new_indefinite_map();
 	g_autoptr(cbor_item_t) item_file = cbor_new_indefinite_map();
 	if (payload->name != NULL) {
-		fu_coswid_firmware_write_tag_string(item_file,
-						    FU_COSWID_TAG_FS_NAME,
-						    payload->name);
+		fu_coswid_write_tag_string(item_file, FU_COSWID_TAG_FS_NAME, payload->name);
 	}
 	if (payload->size != 0) {
-		fu_coswid_firmware_write_tag_uint64(item_file, FU_COSWID_TAG_SIZE, payload->size);
+		fu_coswid_write_tag_u64(item_file, FU_COSWID_TAG_SIZE, payload->size);
 	}
 	if (payload->hashes->len > 0) {
 		g_autoptr(cbor_item_t) item_hashes = cbor_new_indefinite_array();
@@ -715,9 +716,9 @@ fu_coswid_firmware_write_payload(cbor_item_t *root, FuCoswidFirmwarePayload *pay
 			FuCoswidFirmwareHash *hash = g_ptr_array_index(payload->hashes, j);
 			fu_coswid_firmware_write_hash(item_hashes, hash);
 		}
-		fu_coswid_firmware_write_tag_item(item_file, FU_COSWID_TAG_HASH, item_hashes);
+		fu_coswid_write_tag_item(item_file, FU_COSWID_TAG_HASH, item_hashes);
 	}
-	fu_coswid_firmware_write_tag_item(item_payload, FU_COSWID_TAG_FILE, item_file);
+	fu_coswid_write_tag_item(item_payload, FU_COSWID_TAG_FILE, item_file);
 	if (!cbor_array_push(root, item_payload))
 		g_critical("failed to push to indefinite array");
 }
@@ -736,50 +737,41 @@ fu_coswid_firmware_write(FuFirmware *firmware, GError **error)
 	g_autoptr(cbor_item_t) item_meta = cbor_new_indefinite_map();
 
 	/* preallocate the map structure */
-	fu_coswid_firmware_write_tag_string(root, FU_COSWID_TAG_LANG, "en-US");
+	fu_coswid_write_tag_string(root, FU_COSWID_TAG_LANG, "en-US");
 	if (fu_firmware_get_id(firmware) != NULL) {
 		fwupd_guid_t uuid = {0};
 		if (fwupd_guid_from_string(fu_firmware_get_id(firmware),
 					   &uuid,
 					   FWUPD_GUID_FLAG_NONE,
 					   NULL)) {
-			fu_coswid_firmware_write_tag_bytestring(root,
-								FU_COSWID_TAG_TAG_ID,
-								(const guint8 *)&uuid,
-								sizeof(uuid));
+			fu_coswid_write_tag_bytestring(root,
+						       FU_COSWID_TAG_TAG_ID,
+						       (const guint8 *)&uuid,
+						       sizeof(uuid));
 		} else {
-			fu_coswid_firmware_write_tag_string(root,
-							    FU_COSWID_TAG_TAG_ID,
-							    fu_firmware_get_id(firmware));
+			fu_coswid_write_tag_string(root,
+						   FU_COSWID_TAG_TAG_ID,
+						   fu_firmware_get_id(firmware));
 		}
 	}
-	fu_coswid_firmware_write_tag_bool(root, FU_COSWID_TAG_CORPUS, TRUE);
-	if (priv->product != NULL) {
-		fu_coswid_firmware_write_tag_string(root,
-						    FU_COSWID_TAG_SOFTWARE_NAME,
-						    priv->product);
-	}
+	fu_coswid_write_tag_bool(root, FU_COSWID_TAG_CORPUS, TRUE);
+	if (priv->product != NULL)
+		fu_coswid_write_tag_string(root, FU_COSWID_TAG_SOFTWARE_NAME, priv->product);
 	if (fu_firmware_get_version(firmware) != NULL) {
-		fu_coswid_firmware_write_tag_string(root,
-						    FU_COSWID_TAG_SOFTWARE_VERSION,
-						    fu_firmware_get_version(firmware));
-	}
-	if (priv->version_scheme != FU_COSWID_VERSION_SCHEME_UNKNOWN) {
-		fu_coswid_firmware_write_tag_uint16(root,
-						    FU_COSWID_TAG_VERSION_SCHEME,
-						    priv->version_scheme);
-	}
-	fu_coswid_firmware_write_tag_item(root, FU_COSWID_TAG_SOFTWARE_META, item_meta);
-	fu_coswid_firmware_write_tag_string(item_meta, FU_COSWID_TAG_GENERATOR, PACKAGE_NAME);
-	if (priv->summary != NULL) {
-		fu_coswid_firmware_write_tag_string(item_meta,
-						    FU_COSWID_TAG_SUMMARY,
-						    priv->summary);
+		fu_coswid_write_tag_string(root,
+					   FU_COSWID_TAG_SOFTWARE_VERSION,
+					   fu_firmware_get_version(firmware));
 	}
+	if (priv->version_scheme != FU_COSWID_VERSION_SCHEME_UNKNOWN)
+		fu_coswid_write_tag_u16(root, FU_COSWID_TAG_VERSION_SCHEME, priv->version_scheme);
+	fu_coswid_write_tag_item(root, FU_COSWID_TAG_SOFTWARE_META, item_meta);
+	fu_coswid_write_tag_string(item_meta, FU_COSWID_TAG_GENERATOR, PACKAGE_NAME);
+	if (priv->summary != NULL)
+		fu_coswid_write_tag_string(item_meta, FU_COSWID_TAG_SUMMARY, priv->summary);
 	if (priv->colloquial_version != NULL) {
-		fu_coswid_firmware_write_tag_string(item_meta,
-						    FU_COSWID_TAG_COLLOQUIAL_VERSION,
-						    priv->colloquial_version);
+		fu_coswid_write_tag_string(item_meta,
+					   FU_COSWID_TAG_COLLOQUIAL_VERSION,
+					   priv->colloquial_version);
 	}
 
 	/* add entities */
@@ -790,28 +782,27 @@ fu_coswid_firmware_write(FuFirmware *firmware, GError **error)
 			g_autoptr(cbor_item_t) item_entity = cbor_new_indefinite_map();
 			g_autoptr(cbor_item_t) item_roles = cbor_new_indefinite_array();
 			if (entity->name != NULL) {
-				fu_coswid_firmware_write_tag_string(item_entity,
-								    FU_COSWID_TAG_ENTITY_NAME,
-								    entity->name);
+				fu_coswid_write_tag_string(item_entity,
+							   FU_COSWID_TAG_ENTITY_NAME,
+							   entity->name);
 			}
 			if (entity->regid != NULL) {
-				fu_coswid_firmware_write_tag_string(item_entity,
-								    FU_COSWID_TAG_REG_ID,
-								    entity->regid);
+				fu_coswid_write_tag_string(item_entity,
+							   FU_COSWID_TAG_REG_ID,
+							   entity->regid);
 			}
-			for (guint j = 0; entity->roles[j] != FU_COSWID_ENTITY_ROLE_UNKNOWN; j++) {
-				g_autoptr(cbor_item_t) item_role =
-				    cbor_build_uint8(entity->roles[j]);
-				if (!cbor_array_push(item_roles, item_role))
-					g_critical("failed to push to indefinite array");
+			for (guint j = 0; j < FU_COSWID_ENTITY_ROLE_LAST; j++) {
+				if (entity->roles & (1u << j)) {
+					g_autoptr(cbor_item_t) item_role = cbor_build_uint8(j);
+					if (!cbor_array_push(item_roles, item_role))
+						g_critical("failed to push to indefinite array");
+				}
 			}
-			fu_coswid_firmware_write_tag_item(item_entity,
-							  FU_COSWID_TAG_ROLE,
-							  item_roles);
+			fu_coswid_write_tag_item(item_entity, FU_COSWID_TAG_ROLE, item_roles);
 			if (!cbor_array_push(item_entities, item_entity))
 				g_critical("failed to push to indefinite array");
 		}
-		fu_coswid_firmware_write_tag_item(root, FU_COSWID_TAG_ENTITY, item_entities);
+		fu_coswid_write_tag_item(root, FU_COSWID_TAG_ENTITY, item_entities);
 	}
 
 	/* add links */
@@ -821,15 +812,15 @@ fu_coswid_firmware_write(FuFirmware *firmware, GError **error)
 			FuCoswidFirmwareLink *link = g_ptr_array_index(priv->links, i);
 			g_autoptr(cbor_item_t) item_link = cbor_new_indefinite_map();
 			if (link->href != NULL) {
-				fu_coswid_firmware_write_tag_string(item_link,
-								    FU_COSWID_TAG_HREF,
-								    link->href);
+				fu_coswid_write_tag_string(item_link,
+							   FU_COSWID_TAG_HREF,
+							   link->href);
 			}
-			fu_coswid_firmware_write_tag_int8(item_link, FU_COSWID_TAG_REL, link->rel);
+			fu_coswid_write_tag_s8(item_link, FU_COSWID_TAG_REL, link->rel);
 			if (!cbor_array_push(item_links, item_link))
 				g_critical("failed to push to indefinite array");
 		}
-		fu_coswid_firmware_write_tag_item(root, FU_COSWID_TAG_LINK, item_links);
+		fu_coswid_write_tag_item(root, FU_COSWID_TAG_LINK, item_links);
 	}
 
 	/* add payloads */
@@ -839,7 +830,7 @@ fu_coswid_firmware_write(FuFirmware *firmware, GError **error)
 			FuCoswidFirmwarePayload *payload = g_ptr_array_index(priv->payloads, i);
 			fu_coswid_firmware_write_payload(item_payloads, payload);
 		}
-		fu_coswid_firmware_write_tag_item(root, FU_COSWID_TAG_PAYLOAD, item_payloads);
+		fu_coswid_write_tag_item(root, FU_COSWID_TAG_PAYLOAD, item_payloads);
 	}
 
 	/* serialize */
@@ -866,7 +857,6 @@ fu_coswid_firmware_build_entity(FuCoswidFirmware *self, XbNode *n, GError **erro
 {
 	FuCoswidFirmwarePrivate *priv = GET_PRIVATE(self);
 	const gchar *tmp;
-	guint entity_role_cnt = 0;
 	FuCoswidEntityRole role;
 	g_autoptr(GPtrArray) roles = NULL;
 	g_autoptr(FuCoswidFirmwareEntity) entity = g_new0(FuCoswidFirmwareEntity, 1);
@@ -888,7 +878,8 @@ fu_coswid_firmware_build_entity(FuCoswidFirmware *self, XbNode *n, GError **erro
 			XbNode *c = g_ptr_array_index(roles, i);
 			tmp = xb_node_get_text(c);
 			role = fu_coswid_entity_role_from_string(tmp);
-			if (role == FU_COSWID_ENTITY_ROLE_UNKNOWN) {
+			if (role == FU_COSWID_ENTITY_ROLE_UNKNOWN ||
+			    role >= FU_COSWID_ENTITY_ROLE_LAST) {
 				g_set_error(error,
 					    G_IO_ERROR,
 					    G_IO_ERROR_INVALID_DATA,
@@ -896,14 +887,7 @@ fu_coswid_firmware_build_entity(FuCoswidFirmware *self, XbNode *n, GError **erro
 					    tmp);
 				return FALSE;
 			}
-			if (entity_role_cnt >= G_N_ELEMENTS(entity->roles)) {
-				g_set_error_literal(error,
-						    G_IO_ERROR,
-						    G_IO_ERROR_INVALID_DATA,
-						    "too many roles");
-				return FALSE;
-			}
-			entity->roles[entity_role_cnt++] = role;
+			entity->roles |= 1u << role;
 		}
 	}
 
@@ -1123,11 +1107,12 @@ fu_coswid_firmware_export(FuFirmware *firmware, FuFirmwareExportFlags flags, XbB
 		g_autoptr(XbBuilderNode) bc = xb_builder_node_insert(bn, "entity", NULL);
 		fu_xmlb_builder_insert_kv(bc, "name", entity->name);
 		fu_xmlb_builder_insert_kv(bc, "regid", entity->regid);
-		for (guint j = 0; entity->roles[j] != FU_COSWID_ENTITY_ROLE_UNKNOWN; j++) {
-			fu_xmlb_builder_insert_kv(
-			    bc,
-			    "role",
-			    fu_coswid_entity_role_to_string(entity->roles[j]));
+		for (guint j = 0; j < FU_COSWID_ENTITY_ROLE_LAST; j++) {
+			if (entity->roles & (1u << j)) {
+				fu_xmlb_builder_insert_kv(bc,
+							  "role",
+							  fu_coswid_entity_role_to_string(j));
+			}
 		}
 	}
 }
@@ -1171,6 +1156,13 @@ fu_coswid_firmware_class_init(FuCoswidFirmwareClass *klass)
 	klass_firmware->build = fu_coswid_firmware_build;
 	klass_firmware->export = fu_coswid_firmware_export;
 	klass_firmware->get_checksum = fu_coswid_firmware_get_checksum;
+
+#ifdef HAVE_CBOR_SET_ALLOCS
+	/* limit for protection, yes; this is global and there is no context */
+	cbor_set_allocs(fu_coswid_firmware_malloc,
+			fu_coswid_firmware_realloc,
+			fu_coswid_firmware_free);
+#endif
 }
 
 /**
